import logging
import urllib3

from pathlib import Path
from urllib.parse import urlencode

from scrapy.spiders import Spider
from scrapy.signals import response_received, spider_error, item_dropped
from scrapy.http.request import Request as BaseRequest

from gargantext.utils.json import json_dumps
from gargantext.utils.dates import datetime
from gargantext.utils.convert import to_int, to_bool, to_str

from .responses import TextResponse, HtmlResponse, XmlResponse, JsonResponse, \
                       RISResponse


__all__ = ['Scraper', 'Request', 'TextResponse', 'HtmlResponse', 'XmlResponse',
           'JsonResponse', 'RISResponse']


class Request(BaseRequest):
    def __init__(self, url, callback=None, method='GET', headers=None, body=None,
             cookies=None, meta=None, encoding='utf-8', priority=0,
             dont_filter=False, errback=None, flags=None, params=None):

        if params:
            url += '?' + urlencode(params)

        super().__init__(url, callback, method, headers, body, cookies, meta,
                encoding, priority, dont_filter, errback, flags)


class Scraper(Spider):
    MAX_COUNT = 1000
    BATCH_SIZE = 100
    DEBUG_DIR = '/tmp'
    ARGUMENTS = {
        'user':         (to_str, None),
        'corpus':       (to_int, None),
        'callback':     (to_str, None),
        'url':          (to_str, None),
        'limit':        (to_int, None),
        'query':        (to_str, None),
        'count_only':   (to_bool, False),
        'report_every': (to_int, BATCH_SIZE),
    }

    def __init__(self, *args, **kwargs):
        # The default __init__ method take any spider arguments and copy them
        # to the spider as attributes: filter arguments for security purposes,
        # converting them safely and falling back to default when needed.
        spider_args = {
            k: convert(kwargs.get(k)) or default
            for k, (convert, default) in self.ARGUMENTS.items()
        }

        super().__init__(*args, **spider_args)

        self.logger.info("Arguments: %r", spider_args)

        default_parser = getattr(self, 'default_parser', None)
        if default_parser and not hasattr(self, 'parse'):
            # XXX Use setattr to bypass pylint warning...
            setattr(self, 'parse', getattr(self, default_parser))

        self._total = None
        self.counter = 0
        self.status = {"succeeded": 0, "failed": 0, "remaining": 0}
        self.events = []
        self.events_history = []

        # For errors/events reporting
        self.http = urllib3.PoolManager()

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = super().from_crawler(crawler, *args, **kwargs)

        crawler.signals.connect(spider.on_error, signal=spider_error)
        crawler.signals.connect(spider.on_item_dropped, signal=item_dropped)
        crawler.signals.connect(spider.trace, signal=response_received)

        spider.report('event', events=[{
            "level": "INFO",
            "message": "Scraper %s started" % spider.name,
        }])

        return spider

    def start_requests(self):
        if self.url and self.url.startswith('file:'):
            yield self.request(self.url)
        else:
            yield from self.dispatch()

    def request(self, url, callback=None, method='GET', headers=None,
             body=None, cookies=None, meta=None, encoding='utf-8', priority=0,
             dont_filter=False, errback=None, flags=None, params=None):

        if callback is None:
            # Callback has to be specify explicitly to set an errback, so
            # get `parse` method or no-op by default.
            callback = getattr(self, 'parse', lambda x: None)

        if errback is None:
            errback = self.errback

        return Request(url, callback, method, headers, body, cookies, meta,
                encoding, priority, dont_filter, errback, flags, params)

    def reach_limit(self):
        self.counter += 1
        return self.count is None or self.counter >= self.count

    @property
    def logger_name(self):
        return 'scrapers.%s' % self.name

    @property
    def logger(self):
        logger = logging.getLogger(self.logger_name)
        return logging.LoggerAdapter(logger, {'spider': self})

    @property
    def total(self):
        return self._total

    @total.setter
    def total(self, value):
        self.logger.info("Total document count: %s", value)
        self._total = to_int(value)

    @property
    def count(self):
        limit = self.limit or self.MAX_COUNT

        if self.total is None:
            return limit

        return None if limit is None else min(limit, self.total)

    def add_event(self, level, message):
        levels = set(["CRITICAL", "FATAL", "ERROR", "WARNING", "INFO", "DEBUG"])

        level_code = getattr(logging, level, 0) if level in levels else 0
        self.logger.log(level_code, message)

        self.events.append({
            "level": level,
            "message": message,
        })
        self.events_history.extend(self.events)

    def report(self, path, succeeded=0, failed=0, remaining=None, events=None):
        # Silently fail if not bounded to a crawler or callback is undefined
        if not getattr(self, 'crawler', None) or not self.callback:
            return

        assert path in ('output', 'error', 'event')

        self.status.update(succeeded=succeeded,
                           failed=failed,
                           remaining=remaining)

        events = self.events + (events or [])
        self.events = []

        url = '%s/%s' % (self.callback, path)
        status = dict(events=events, **self.status)
        data = { k: v for k, v in status.items() if v is not None }

        try:
            self.http.request('POST', url,
                body=json_dumps(data).encode('utf-8'),
                headers={'Content-Type': 'application/json'})
        except urllib3.exceptions.HTTPError:
            # Ignore HTTP errors when reporting...
            pass

    def on_error(self, failure, response, spider):
        # FIXME Any reporting to do here?
        self.logger.debug("Catched error: %r %r %r", failure, response, spider)

    def on_item_dropped(self, item, response, exception, spider):
        reason = ' '.join(exception.args)
        self.add_event("ERROR", "Ignore document: %s" % reason)

    def errback(self, failure):
        exc = failure.value
        response = getattr(exc, 'response', None)
        message = "Error in %s scraper" % self.name

        if response is not None:
            message = "%s while fetching %s, got HTTP %s" % (
                message, response.request.url, response.status)

        message = "%s: %s" % (message, failure.getErrorMessage())

        self.add_event("FATAL", message)

    def trace(self, response, request, spider):
        content_type = response.headers.get('content-type', b'').decode()

        self.logger.info('Content-Type=%s; type(response)=%s;',
            content_type, type(response).__name__)

        path = Path(self.DEBUG_DIR).absolute()
        date = datetime.now().strftime("%Y%m%d_%H%m_%s.%f")
        ext = '.html' if isinstance(response, HtmlResponse) else \
              '.xml' if isinstance(response, XmlResponse) else \
              '.json' if isinstance(response, JsonResponse) else \
              '.txt' if isinstance(response, TextResponse) else \
              ''

        filename = '%s-%s%s' % (spider.logger_name, date, ext)
        filepath = str(path / filename)

        with open(filepath, 'wb') as f:
            f.write(response.body)
