import logging

from scrapy.loader import ItemLoader

from gargantext.utils.dates import DEFAULT_DATETIME, to_datetime
from gargantext.utils.lang import lang_name

from .base import TypedItem
from .fields import DateTime, String, Int, Real, Dict, List, Either, \
                    DropDocument


logger = logging.getLogger(__name__)


class Document(TypedItem):
    id = String(required=True)
    title = String(required=True, default='')
    abstract = String(required=True, default='')
    source = String(required=True, default='')
    authors = List(Either(String, Dict), required=True, default=[])
    url = String(required=True, default='')
    publication = DateTime(required=True, default=DEFAULT_DATETIME)

    lang = String
    lang_detected = String
    lang_detected_confidence = Real

    creation = DateTime
    type = String
    doi = String
    isbn = String
    issue = String

    struct_inst = List(Int)
    struct_dept = List(Int)
    struct_labs = List(Int)
    struct_team = List(Int)

    publication_year = Int
    publication_month = Int
    publication_day = Int

    creation_year = Int
    creation_month = Int
    creation_day = Int

    def preprocess(self):
        warnings = super().preprocess() or []

        for field_name in ['publication', 'creation']:
            if field_name in self:
                raw = self.get(field_name)
                default = self.fields[field_name].get('default')
                required = self.fields[field_name].get('required')

                try:
                    value = to_datetime(raw, default=default) if raw else default
                except ValueError as e:
                    value = default if required else None
                    warnings.append("%s: %s" % (field_name, ' '.join(e.args)))

                self[field_name] = value

                if value:
                    self[field_name + '_year']  = int(value.strftime("%Y"))
                    self[field_name + '_month'] = int(value.strftime("%m"))
                    self[field_name + '_day']   = int(value.strftime("%d"))

        self['lang'] = lang_name(self.get('lang'))

        return warnings

    def validate(self):
        warnings = []

        if not self.get('title'):
            warnings.append('No title, is something wrong?')

        if not self.get('title') and not self.get('abstract'):
            raise DropDocument('Missing both title and abstract')

        return warnings + (super().validate() or [])


class DocumentLoader(ItemLoader):
    default_item_class = Document

    def __init__(self, selector, *args, **kwargs):
        kwargs['selector'] = selector
        super().__init__(*args, **kwargs)

    def add_xpaths_text(self, xpaths):
        for field_name, xpath in xpaths.items():
            self.add_xpath(field_name, '%s/text()' % xpath)

    def add_values(self, values):
        self.add_value(None, values)

    def parse(self, obj):
        return NotImplementedError("don't use DocumentLoader directly.")

    def load(self):
        self.parse(self.selector)
        return self.load_item()
