import logging

from datetime import datetime

from scrapy.item import Field

from ..exceptions import DropDocument


__all__ = ['FieldType', 'ListType', 'EitherType', 'Any', 'DateTime', 'String',
           'Int', 'Real', 'List', 'Dict', 'Either', 'DropDocument']


logger = logging.getLogger(__name__)


class FieldType(Field):
    def __init__(self, name, *args, **kwargs):
        self.__name__ = name
        self.args = list(args)
        super().__init__(**kwargs)

    def check_type(self, instance):
        typ = self.get('type')
        return isinstance(instance, typ) if typ else True

    def __call__(self, *args, **kwargs):
        return type(self)(self.__name__, *args, **kwargs)

    def __instancecheck__(self, instance):
        return self.check_type(instance) or \
                (not self.get('required') and instance is None)

    def __repr__(self):
        if not self.args and not self:
            return self.__name__

        args = ', '.join(str(a) for a in self.args)
        opts = ', '.join(('%s=%s' % (k, getattr(v, '__name__', repr(v)))
                         for k, v in self.items()))

        params = '%s, %s' % (args, opts) if args and opts else args or opts

        return '%s(%s)' % (self.__name__, params)


class ListType(FieldType):
    def __init__(self, name, item_type, *args, **kwargs):
        kwargs.setdefault('item_type', item_type)
        super().__init__(name, *args, **kwargs)
        self.args = [item_type] + self.args

    def check_type(self, instance):
        item_type = self.get('item_type')
        return (isinstance(instance, list) and
                all(isinstance(item, item_type) for item in instance))


class EitherType(FieldType):
    def check_type(self, instance):
        return any(isinstance(instance, t) for t in self.args)


Any = FieldType('Any')
DateTime = FieldType('DateTime', type=datetime)
String = FieldType('String', type=str)
Int = FieldType('Int', type=int)
Real = FieldType('Real', type=float)
Dict = FieldType('Dict', type=dict)
List = ListType('List', Any)
Either = EitherType('Either')
