# XXX This is needed because database and models depends on Django
_DJANGO_SETUP = False
if not _DJANGO_SETUP:
    import django
    django.setup()
    _DJANGO_SETUP = True

import logging

from gargantext.models.nodes import DocumentNode, CorpusNode
from gargantext.core.db import Session
from gargantext.utils.dates import datetime
from gargantext.utils.lang import lang_detect


logger = logging.getLogger(__name__)


class PreprocessPipeline(object):
    def add_warnings(self, scraper, warnings):
        if warnings:
            for warn in warnings:
                scraper.add_event("WARNING", warn)

    def process_item(self, item, scraper):
        w = item.preprocess()
        self.add_warnings(scraper, w)
        return item


class ValidationPipeline(PreprocessPipeline):
    def process_item(self, item, scraper):
        w = item.validate()
        self.add_warnings(scraper, w)
        return item


class LangDetectPipeline(object):
    THRESHOLD = 0

    def process_item(self, item, scraper):
        parts = [item.get('title'), item.get('abstract')]
        text = ' '.join((p or '' for p in parts)).strip()

        if text:
            lang, confidence = lang_detect(text)
            if confidence >= self.THRESHOLD:
                item['lang_detected'] = lang
                item['lang_detected_confidence'] = confidence
            else:
                logger.warning(
                    "Detected %s on %r with insufficient confidence of %s",
                    lang, (text[:75] + '..') if len(text) > 75 else text,
                    confidence)

        return item


class DatabasePipeline(object):
    def open_spider(self, scraper):
        if not scraper.user:
            raise ValueError("No user specified")

        if not scraper.corpus:
            raise ValueError("No corpus specified")

        self.db = Session()
        self.db.login(scraper.user)
        self.corpus = self.db.query(CorpusNode).filter_by(id=scraper.corpus).one_or_none()

    def close_spider(self, scraper):
        resources = self.corpus.hyperdata.get('resources', [])
        resources.append({
            "date": datetime.now(),
            # TODO Raw files storage and listing in paths
            "paths": None,
            "scraper": scraper.name,
            "query": scraper.query,
            "status": scraper.status,
            "events": scraper.events_history,
        })

        self.corpus['resources'] = resources
        self.corpus.save_hyperdata()

        self.db.commit()

    def process_item(self, item, scraper):
        doc = DocumentNode(name=item.get('title')[:DocumentNode.NAME_MAXLEN],
                           parent_id=self.corpus.id,
                           hyperdata=dict(item))

        self.corpus.related.append(doc)

        if scraper.counter % scraper.BATCH_SIZE == 0:
            self.db.commit()

        return item


class ReportingPipeline(object):
    def __init__(self, crawler):
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def report(self, path, scraper, completed=False):
        # See https://doc.scrapy.org/en/latest/topics/stats.html
        stats = self.crawler.stats

        # See https://doc.scrapy.org/en/latest/topics/api.html#stats-collector-api
        scraped_count = stats.get_value('item_scraped_count', 0)
        dropped_count = stats.get_value('item_dropped_count', 0)
        pending_count = None if scraper.count is None else \
                        max(0, scraper.count - scraped_count)

        if completed and pending_count is not None and pending_count > 0:
            inflexion = "s were" if pending_count > 1 else " was"
            message = "%s item%s never fetched" % (pending_count, inflexion)

            scraper.add_event("WARNING", message)

            dropped_count += pending_count
            pending_count = 0

        scraper.report(path,
            succeeded=scraped_count,
            failed=dropped_count,
            remaining=pending_count)

    def close_spider(self, scraper):
        self.report('output', scraper, completed=True)

    def process_item(self, item, scraper):
        scraped_count = self.crawler.stats.get_value('item_scraped_count', 0)

        if scraped_count % scraper.report_every == 0:
            self.report('event', scraper)

        return item
